@@ -15,25 +15,25 @@ import pkg_resources
 ### Load in project resources
 # Full word list
 ocrfixr = importlib_resources.files("ocrfixr")
-word_set = (ocrfixr / "data" / "SCOWL_70.txt").read_text().split()
+word_set = (ocrfixr / "data" / "SCOWL_70.txt").read_text(encoding='utf-8').split()
 word_set = set(word_set)
 
 # List of words NOT in SCOWL 70 that we should ignore anyways
-ignore_misreads = (ocrfixr / "data" / "ignore_these_misspells.txt").read_text().split()
+ignore_misreads = (ocrfixr / "data" / "ignore_these_misspells.txt").read_text(encoding='utf-8').split()
 ignore_set_from_pkg = set(ignore_misreads)
 
 # dict of common scannos to check for (bypasses the context check, since these are clear mappings)
-common_scannos = (ocrfixr / "data" / "scannos_common.txt").read_text()
+common_scannos = (ocrfixr / "data" / "scannos_common.txt").read_text(encoding='utf-8')
 common_scannos = ast.literal_eval(common_scannos)
 common = set(common_scannos)
 
-# dict of specifically tricky scannos to check for (bypasses the context check, since these are clear mappings)
-stealth_scannos = (ocrfixr / "data" / "scannos_stealth.txt").read_text()
+# dict of specifically tricky scannos to check for - misspellings that create real words (arid - and)
+stealth_scannos = (ocrfixr / "data" / "scannos_stealth.txt").read_text(encoding='utf-8')
 stealth_scannos = ast.literal_eval(stealth_scannos)
 stealth = set(stealth_scannos)
 
 # dict of OCRfixr suggestions that are known to be bad. This list prevents them from ever being suggested.
-ignore_suggestions = (ocrfixr / "data" / "ignore_these_suggestions.txt").read_text()
+ignore_suggestions = (ocrfixr / "data" / "ignore_these_suggestions.txt").read_text(encoding='utf-8')
 ignore_suggestions = ast.literal_eval(ignore_suggestions)
 #ignore_suggestions = set(ignore_suggestions)
 
@@ -83,13 +83,16 @@ class spellcheck:
         # Also, drop all words with trailing numbers flanked by punctuation, indicating a footnote reference (money.4, item[1]) rather than a misspelling
         # Also, drop any 1-character "words" 
 
-        no_hyphens = re.compile(".*-.*|.*'.*|.*’.*|[0-9]+")
-        no_caps = re.compile('[^A-Z][a-z0-9]{1,}')
+        no_hyphens = re.compile(".*-.*|.*'.*|.*\.{3,}.*|.*’.*|[0-9]+")
+        no_caps = re.compile('[^A-Z]{2,}')
         no_footnotes = re.compile('.*[0-9]{1,}[^A-z]?$')
+        no_roman_nums = re.compile('[xlcviXLCVI.:,-;]+$')
+        no_eth_endings = re.compile('.*eth|.*est$')
+        no_format_tags = re.compile('</?[a-z]>.*|.*</?[a-z]>')
         no_list_items = re.compile('.*\\)|.*\\]')
         all_nums = re.compile('^[0-9]{1,}$')
 
-        words = [x for x in tokens if not no_hyphens.match(x) and no_caps.match(x) and not no_footnotes.match(x) and not no_list_items.match(x)]
+        words = [x for x in tokens if not no_hyphens.match(x) and no_caps.match(x) and not no_footnotes.match(x) and not no_roman_nums.match(x) and not no_eth_endings.match(x) and not no_format_tags.match(x) and not no_list_items.match(x)]
 
         # then, remove punct from each remaining token (such as trailing commas, periods, quotations ('' & ""), but KEEPING contractions). 
         no_punctuation = [l.strip(string.punctuation+"”“’‘") for l in words]
@@ -97,12 +100,30 @@ class spellcheck:
         
         
         # if a word is not in the SCOWL 70 word list, it is assumed to be a misspelling.
-        # also allows for user to specify terms to NOT look at (for example, known slang in the text) = ignore_set_from_user, plus ignore_set_from_pkg, a small set of problematic terms that aren't "words", but are ignored by default (example: th, as in "7 th")
+        unrecognized = []
+        for i in words_to_check:
+            if i not in word_set:
+                unrecognized.append(i)
+        
+        
+        # throw away any paragraphs where > 30% of the words are unrecognized - this makes context-generation spotty AND likely indicates a messy post-script/footnote, or even another language. This limits trigger-happy changes to messy text.
+        L1 = len(tokens)
+        L0 = len(unrecognized)
+        if L0/L1 > 0.30 and L1 > 10:
+            unrecognized = []
+        else: 
+            unrecognized = unrecognized
+      
+        
+        # Allow user to specify terms to NOT look at (for example, known slang in the text) = ignore_set_from_user
+        # plus, remove problematic words = ignore_set_from_pkg, 
+        # this contains a small set of problematic terms that aren't "words" (example: th, as in "7 th")
+        # as well as the 2,000 most-common words in the following languages: Latin, Greek, French, German, Spanish
         ignore_set_from_user = set(self.ignore_words)
         
         misread = []
-        for i in words_to_check:
-            if i not in word_set and i not in ignore_set_from_pkg and i not in ignore_set_from_user:
+        for i in unrecognized:
+            if i not in ignore_set_from_pkg and i not in ignore_set_from_user:
                 misread.append(i)
 
         # add scannos to misreads, if option is selected        
@@ -112,16 +133,6 @@ class spellcheck:
             for i in tokens:
                 if i not in misread and i in common or i in stealth:
                     misread.append(i)
-
-                
-        L0 = len(tokens)
-        L1 = len(misread)
-        
-        # throw away any paragraphs where > 30% of the words are unrecognized - this makes context-generation spotty AND likely indicates a messy post-script/footnote, or even another language. This limits trigger-happy changes to messy text.
-        if L0 < 1:
-            misread = misread
-        elif L1/L0 > 0.30 and L0 > 10:
-            misread = []
             
         return(misread)
 
@@ -255,6 +266,7 @@ class spellcheck:
     def _FIND_REPLACEMENTS(self, misreads):
         SC = [] 
         bert = []
+        punct_split_fixes = {}
         common_scanno_fixes = {}
         
         # for each misread, get all spellcheck suggestions
@@ -267,8 +279,12 @@ class spellcheck:
             # for stealth scannos - these are valid (yet incorrect) words. So, instead of SUGGEST_SPELLCHECK (which would return the same word supplied), take the value from the stealth_scanno dict, which is the desired word to check for in BERT context (arid --> and)
             elif self.common_scannos == "T" and i in stealth:
                 SC.append(stealth_scannos.get(i).split(" "))
-                bert.append(self.__SUGGEST_BERT(text = self.__SET_MASK(i,'[MASK]', self.text), 
-                                                number_to_return = self.top_k))
+                SB = self.__SUGGEST_BERT(text = self.__SET_MASK(i,'[MASK]', self.text), 
+                                                number_to_return = self.top_k)
+                
+                # if the original stealth scanno also makes sense in context, then don't record the suggestion
+                if i not in SB:
+                    bert.append(SB)
             
             # for all other unrecognized words, get all spellcheck suggestions from symspell
             else:
@@ -278,24 +294,49 @@ class spellcheck:
                     # if no spellcheck suggestion given, remove from misreads (ie. dont bother checking BERT context - it won't get used)
                     misreads.delete(i)
                 else: 
-                    # tee up symspell suggestion for comparison to BERT suggestion
-                    SC.append(spellcheck)  
-                    
-                    # For multi-word phrases, the BERT context check will feed in the first word into the text, then confirm whether second word fits the context of the sentence. If so, the two word phrase will be accepted as a valid correction
+                   # For multi-word phrases....
                     if self.suggest_unsplit == "T" and len(spellcheck) == 1 and str(spellcheck).count(' ') == 1:
-                        mw = ''.join(spellcheck)
-                        fw = re.findall("^[^\s]+", mw).pop()
-                        SB = self.__SUGGEST_BERT(text = self.__SET_MASK(i, fw + ' [MASK]', self.text), 
-                                                        number_to_return = self.top_k)   
                         
-                        # Tack the first word onto the results for each BERT context suggestion. These are compared against the multi-word phrase provided by sympell
-                        SBi = []
-                        for x in SB:
-                            SBi.append(fw + ' ' + x)
-                        
-                        bert.append(SBi)
+                        # if the phrase was already separated by a comma or period ("shall.cultivate"), skip the BERT context check
+                        # just confirm that both word halves are valid
+                        if "." in i or "," in i:
+                            mw = ''.join(spellcheck)                           
+                            fw = re.findall("^[^\s,]+", mw).pop()
+                            sw = re.findall("[^\s]+$", mw).pop()
+                            
+                            if fw in word_set and sw in word_set:
+                                # if first letter after period split is uppercased, retain it and add a period ('ended.He' --> 'ended. He')
+                                if len(re.findall("\.{1,}([A-Z][a-z]+)", i)) > 0:
+                                    fw = fw + '. '
+                                    sw = str.title(sw)
+                                    mw = fw + sw
+
+                                add_splitto = {i:mw}
+                                punct_split_fixes.update(add_splitto)
+
+                    
+                        # If symspell has to pick an arbitrary cutoff between the words ("anhour"), check the suggestion using BERT context
+                        # Feed in the first word into the text, then confirm whether second word fits the context of the sentence using BERT. If so, the two word phrase will be accepted as a valid correction
+                        else:
+                            # tee up symspell suggestion for comparison to BERT suggestion
+                            SC.append(spellcheck)  
+                            
+                            mw = ''.join(spellcheck)
+                            fw = re.findall("^[^\s]+", mw).pop()
+                            SB = self.__SUGGEST_BERT(text = self.__SET_MASK(i, fw + ' [MASK]', self.text), 
+                                                            number_to_return = self.top_k)   
+                            
+                            # Tack the first word onto the results for each BERT context suggestion. These are compared against the multi-word phrase provided by sympell
+                            SBi = []
+                            for x in SB:
+                                SBi.append(fw + ' ' + x)
+                            
+                            bert.append(SBi)
+                              
 
                     else:    
+                        SC.append(spellcheck)  
+
                         # otherwise, just mask the misspelled word for BERT context check, which will be compared against symspell
                         bert.append(self.__SUGGEST_BERT(text = self.__SET_MASK(i,'[MASK]', self.text), 
                                                         number_to_return = self.top_k))
@@ -343,6 +384,7 @@ class spellcheck:
           
         # Add the common scannos to the mix
         fixes.update(common_scanno_fixes)
+        fixes.update(punct_split_fixes)
     
     
         # Remove all dict entries where replacement is same as initial problematic text
@@ -456,15 +498,14 @@ class spellcheck:
 
 
 
-
-# TODO - (FULL_PARAGRAPHS) Need to allow BERT context to draw from all lines in a full paragraph (currently resets at each newline -- this corresponds to 1 line of text in a Gutenberg text, and likely leads to degraded spellcheck performance due to loss of context)
 # TODO - (ADD_DICTS) Need to add selectable foreign language dictionaries 
+# TODO - (IGNORE_SPLIT_WORDS) need to ignore the first word of a new page, since these can be split words across pages (this may also just be tied up in the unsplit functionality, where this word should have a leading * to denote a split word)
 # TODO - (ADD_STEALTHOS) Need to add additional common stealth scannos to OCRfixr
+# TODO - (FULL_PARAGRAPHS) Need to allow BERT context to draw from all lines in a full paragraph (currently resets at each newline -- this corresponds to 1 line of text in a Gutenberg text, and likely leads to degraded spellcheck performance due to loss of context)
 # TODO - (GutenBERT) fine-tune BERT model on Gutenberg texts, to improve relatedness of context suggestions
 # TODO - (WARM_UP) can we somehow negate the warm-up time for the transformers unmasker?
     # pipelines = 7 secs
     # symspellpy dictionary load = 3 seconds
-# TODO - (IGNORE_SPLIT_WORDS) need to ignore the first word of a new page, since these can be split words across pages (this may also just be tied up in the unsplit functionality, where this word should have a leading * to denote a split word)
 
 
 # Note: find-replace is not instance-specific, it is paragraph specific..."yov" will be replaced with "you" in all instances found in that section of text. It would be rare, but this may cause issues when a repeated scanno is valid & not valid within the same paragraph
